@@ -1,10 +1,17 @@
-# Evaluate list members that are formulae, using the map data as the environment
-# (if provided, otherwise the formula environment)
-evalFormula = function(list, data) {
-  evalAll = function(x) {
+#' Evaluate list members that are formulae, using the map data as the environment
+#' (if provided, otherwise the formula environment)
+#' @param list with members as formulae
+#' @param data map data
+#' @export
+evalFormula <- function(list, data) {
+  evalAll <- function(x) {
     if (is.list(x)) {
-      structure(lapply(x, evalAll), class = class(x))
-    } else resolveFormula(x, data)
+      # Use `x[] <-` so attributes on x are preserved
+      x[] <- lapply(x, evalAll)
+      x
+    } else {
+      resolveFormula(x, data)
+    }
   }
   evalAll(list)
 }
@@ -15,10 +22,14 @@ evalFormula = function(list, data) {
 # polygon lists (returned from polygonData()) use `attr(x, "bbox")` (though at
 # least they are the same shape as the Spatial bounding boxes).
 
-# Notifies the map of new latitude/longitude of items of interest on the map, so
-# that we can expand the limits (i.e. bounding box). We will use this as the
+#' Notifies the map of new latitude/longitude of items of interest on the map
+# So that we can expand the limits (i.e. bounding box). We will use this as the
 # initial view if the user doesn't explicitly specify bounds using fitBounds.
-expandLimits = function(map, lat, lng) {
+#' @param map map object
+#' @param lat vector of latitudes
+#' @param lng vector of longitudes
+#' @export
+expandLimits <- function(map, lat, lng) {
   if (is.null(map$x$limits)) map$x$limits = list()
 
   # We remove NA's and check the lengths so we never call range() with an empty
@@ -33,9 +44,12 @@ expandLimits = function(map, lat, lng) {
   map
 }
 
-# Same as expandLimits, but takes a polygon (that presumably has a bbox attr)
-# rather than lat/lng.
-expandLimitsBbox = function(map, poly) {
+#' Same as expandLimits, but takes a polygon (that presumably has a bbox attr)
+#' rather than lat/lng.
+#' @param map map object
+#' @param poly A spatial object representing a polygon.
+#' @export
+expandLimitsBbox <- function(map, poly) {
   bbox = attr(poly, "bbox", exact = TRUE)
   if (is.null(bbox)) stop("Polygon data had no bbox")
   expandLimits(map, bbox[2, ], bbox[1, ])
@@ -47,7 +61,7 @@ bboxNull = cbind(min = c(x = Inf, y = Inf), max = c(x = -Inf, y = -Inf))
 
 # Combine two bboxes; the result will use the mins of the mins and the maxes of
 # the maxes.
-bboxAdd = function(a, b) {
+bboxAdd <- function(a, b) {
   cbind(
     min = pmin(a[, 1], b[, 1]),
     max = pmax(a[, 2], b[, 2])
@@ -74,13 +88,13 @@ clearGroup <- function(map, group) {
 #'   groups interactively
 #'
 #' @export
-showGroup = function(map, group) {
+showGroup <- function(map, group) {
   invokeMethod(map, getMapData(map), 'showGroup', group)
 }
 
 #' @rdname showGroup
 #' @export
-hideGroup = function(map, group) {
+hideGroup <- function(map, group) {
   invokeMethod(map, getMapData(map), 'hideGroup', group)
 }
 
@@ -100,9 +114,9 @@ hideGroup = function(map, group) {
 #'   \url{http://leafletjs.com/reference.html}
 #' @describeIn map-layers Add a tile layer to the map
 #' @export
-addTiles = function(
+addTiles <- function(
   map,
-  urlTemplate = 'http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
+  urlTemplate = '//{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
   attribution = NULL,
   layerId = NULL,
   group = NULL,
@@ -174,7 +188,7 @@ epsg3857 <- "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y
 #' leaflet() %>% addTiles() %>%
 #'   addRasterImage(r, colors = "Spectral", opacity = 0.8)
 #' @export
-addRasterImage = function(
+addRasterImage <- function(
   map,
   x,
   colors = "Spectral",
@@ -224,13 +238,13 @@ projectRasterForLeaflet <- function(x) {
 
 #' @rdname remove
 #' @export
-removeImage = function(map, layerId) {
+removeImage <- function(map, layerId) {
   invokeMethod(map, NULL, 'removeImage', layerId)
 }
 
 #' @rdname remove
 #' @export
-clearImages = function(map) {
+clearImages <- function(map) {
   invokeMethod(map, NULL, 'clearImages')
 }
 
@@ -242,9 +256,10 @@ clearImages = function(map) {
 #' minZoom,maxZoom,maxNativeZoom,tileSize,subdomains,errorTileUrl,tms,continuousWorld,noWrap,zoomOffset,zoomReverse,zIndex,unloadInvisibleTiles,updateWhenIdle,detectRetina,reuseTiles
 #' the tile layer options; see
 #' \url{http://leafletjs.com/reference.html#tilelayer}
+#' @param ... extra options passed to underlying Javascript object constructor.
 #' @describeIn map-options Options for tile layers
 #' @export
-tileOptions = function(
+tileOptions <- function(
   minZoom = 0,
   maxZoom = 18,
   maxNativeZoom = NULL,
@@ -261,8 +276,8 @@ tileOptions = function(
   unloadInvisibleTiles = NULL,
   updateWhenIdle = NULL,
   detectRetina = FALSE,
-  reuseTiles = FALSE
-  # bounds = TODO
+  reuseTiles = FALSE,
+  ...
 ) {
   list(
     minZoom = minZoom, maxZoom = maxZoom, maxNativeZoom = maxNativeZoom,
@@ -271,7 +286,8 @@ tileOptions = function(
     zoomOffset = zoomOffset, zoomReverse = zoomReverse, opacity = opacity,
     zIndex = zIndex, unloadInvisibleTiles = unloadInvisibleTiles,
     updateWhenIdle = updateWhenIdle, detectRetina = detectRetina,
-    reuseTiles = reuseTiles
+    reuseTiles = reuseTiles,
+    ...
   )
 }
 
@@ -301,13 +317,13 @@ tileOptions = function(
 #'
 #' @name remove
 #' @export
-removeTiles = function(map, layerId) {
+removeTiles <- function(map, layerId) {
   invokeMethod(map, getMapData(map), 'removeTiles', layerId)
 }
 
 #' @rdname remove
 #' @export
-clearTiles = function(map) {
+clearTiles <- function(map) {
   invokeMethod(map, NULL, 'clearTiles')
 }
 
@@ -316,7 +332,7 @@ clearTiles = function(map) {
 #' @param layers comma-separated list of WMS layers to show
 #' @describeIn map-layers Add a WMS tile layer to the map
 #' @export
-addWMSTiles = function(
+addWMSTiles <- function(
   map, baseUrl, layerId = NULL, group = NULL,
   options = WMSTileOptions(), attribution = NULL, layers = ''
 ) {
@@ -331,13 +347,12 @@ addWMSTiles = function(
 #' @param transparent if \code{TRUE}, the WMS service will return images with
 #'   transparency
 #' @param version version of the WMS service to use
-#' @param crs Coordinate Reference System to use for the WMS requests, defaults
+#' @param crs Coordinate Reference System to use for the WMS requests, defaults.
+#' @seealso \code{\link{leafletCRS}}
 #'   to map CRS (don't change this if you're not sure what it means)
-#' @param ... other tile options for \code{WMSTileOptions()} (all arguments of
-#'   \code{tileOptions()} can be used)
 #' @describeIn map-options Options for WMS tile layers
 #' @export
-WMSTileOptions = function(
+WMSTileOptions <- function(
   styles = '', format = 'image/jpeg', transparent = FALSE, version = '1.1.1',
   crs = NULL, ...
 ) {
@@ -357,6 +372,7 @@ WMSTileOptions = function(
 #'   the latitude column from \code{data})
 #' @param popup a character vector of the HTML content for the popups (you are
 #'   recommended to escape the text using \code{\link[htmltools]{htmlEscape}()}
+#' @param popupOptions A Vector of \code{\link{popupOptions}} to provide popups
 #'   for security reasons)
 #' @param layerId the layer id
 #' @param group the name of the group the newly created layers should belong to
@@ -369,7 +385,7 @@ WMSTileOptions = function(
 #'   initially, but can be overridden
 #' @describeIn map-layers Add popups to the map
 #' @export
-addPopups = function(
+addPopups <- function(
   map, lng = NULL, lat = NULL, popup, layerId = NULL, group = NULL,
   options = popupOptions(),
   data = getMapData(map)
@@ -385,7 +401,7 @@ addPopups = function(
 #' popup options; see \url{http://leafletjs.com/reference.html#popup}
 #' @describeIn map-options Options for popups
 #' @export
-popupOptions = function(
+popupOptions <- function(
   maxWidth = 300,
   minWidth = 50,
   maxHeight = NULL,
@@ -398,44 +414,103 @@ popupOptions = function(
   # autoPanPadding = TODO,
   zoomAnimation = TRUE,
   closeOnClick = NULL,
-  className = ""
+  className = "",
+  ...
 ) {
   list(
     maxWidth = maxWidth, minWidth = minWidth, maxHeight = maxHeight,
     autoPan = autoPan, keepInView = keepInView, closeButton = closeButton,
-    zoomAnimation = zoomAnimation, closeOnClick = closeOnClick, className = className
+    zoomAnimation = zoomAnimation, closeOnClick = closeOnClick, className = className, ...
   )
 }
 
 #' @rdname remove
 #' @export
-removePopup = function(map, layerId) {
+removePopup <- function(map, layerId) {
   invokeMethod(map, getMapData(map), 'removePopup', layerId)
 }
 
 #' @rdname remove
 #' @export
-clearPopups = function(map) {
+clearPopups <- function(map) {
   invokeMethod(map, NULL, 'clearPopups')
 }
 
+#' Sanitize textual labels
+#'
+#' This is a helper function used internally to HTML-escape user-provided
+#' labels. It prevents strings from unintentionally being treated as HTML when
+#' they are intended to be plaintext.
+#'
+#' @param label A vector or list of plain characters or HTML (marked by
+#'   \code{\link[htmltools]{HTML}}), or a formula that resolves to such a value.
+#' @param data A data frame over which the formua is evaluated.
+#'
+#' @keywords internal
+#' @export
+safeLabel <- function(label, data) {
+  if (is.null(label)) {
+    return(label)
+  }
+
+  label <- evalFormula(label, data)
+  if(! (inherits(label, "html") ||
+                    sum(sapply(label,function(x){!inherits(x,'html')})) == 0)) {
+    label <- htmltools::htmlEscape(label)
+  }
+  label
+}
+
+#' @param
+#' noHide,direction,offset,textsize,textOnly,style
+#' label options; see \url{https://github.com/Leaflet/Leaflet.label#options}
+#' @describeIn map-options Options for labels
+#' @export
+labelOptions <- function(
+  clickable = FALSE,
+  noHide = FALSE,
+  className = '',
+  direction = 'right',
+  #pane = NULL,
+  offset = c(12,-15),
+  opacity = 1,
+  textsize = "10px",
+  textOnly = FALSE,
+  style = NULL,
+  zoomAnimation = TRUE,
+  ...
+) {
+  list(
+    clickable = clickable, noHide = noHide, direction = direction,
+    opacity = opacity, offset = offset,
+    textsize = textsize, textOnly = textOnly, style = style,
+    zoomAnimation = zoomAnimation, className = className, ...
+  )
+}
+
 #' @param icon the icon(s) for markers; an icon is represented by an R list of
 #'   the form \code{list(iconUrl = '?', iconSize = c(x, y))}, and you can use
 #'   \code{\link{icons}()} to create multiple icons; note when you use an R list
 #'   that contains images as local files, these local image files will be base64
 #'   encoded into the HTML page so the icon images will still be available even
 #'   when you publish the map elsewhere
+#' @param label a character vector of the HTML content for the labels
+#' @param labelOptions A Vector of \code{\link{labelOptions}} to provide label
+#' options for each label. Default \code{NULL}
 #' @param clusterOptions if not \code{NULL}, markers will be clustered using
 #'   \href{https://github.com/Leaflet/Leaflet.markercluster}{Leaflet.markercluster};
 #'    you can use \code{\link{markerClusterOptions}()} to specify marker cluster
 #'   options
 #' @param clusterId the id for the marker cluster layer
-#' @describeIn map-layers Add markders to the map
+#' @describeIn map-layers Add markers to the map
 #' @export
-addMarkers = function(
+addMarkers <- function(
   map, lng = NULL, lat = NULL, layerId = NULL, group = NULL,
   icon = NULL,
   popup = NULL,
+  popupOptions = NULL,
+  label = NULL,
+  labelOptions = NULL,
   options = markerOptions(),
   clusterOptions = NULL,
   clusterId = NULL,
@@ -468,18 +543,56 @@ addMarkers = function(
 
   pts = derivePoints(data, lng, lat, missing(lng), missing(lat), "addMarkers")
   invokeMethod(
-    map, data, 'addMarkers', pts$lat, pts$lng, icon, layerId, group, options, popup,
-    clusterOptions, clusterId
+    map, data, 'addMarkers', pts$lat, pts$lng, icon, layerId, group, options,
+    popup, popupOptions, clusterOptions, clusterId,
+    safeLabel(label, data), labelOptions,
+    getCrosstalkOptions(data)
   ) %>% expandLimits(pts$lat, pts$lng)
 }
 
-markerClusterDependencies = function() {
+getCrosstalkOptions <- function(data) {
+  if (is.SharedData(data)) {
+    list(ctKey = data$key(), ctGroup = data$groupName())
+  } else {
+    NULL
+  }
+}
+
+#' @describeIn map-layers Add Label only markers to the map
+#' @export
+addLabelOnlyMarkers <- function(
+  map, lng = NULL, lat = NULL, layerId = NULL, group = NULL,
+  icon = NULL,
+  label = NULL,
+  labelOptions = NULL,
+  options = markerOptions(),
+  clusterOptions = NULL,
+  clusterId = NULL,
+  data = getMapData(map)
+) {
+  do.call(addMarkers, filterNULL(list(
+    map = map, lng = lng, lat = lat, layerId = layerId,
+    group = group,
+    icon = makeIcon(
+      iconUrl = system.file('htmlwidgets/lib/leaflet/images/1px.png', package='leaflet'),
+      iconWidth = 1, iconHeight = 1),
+      label = label,
+      labelOptions = labelOptions,
+      options = options,
+      clusterOptions = clusterOptions,
+      clusterId = clusterId,
+      data = data
+  )))
+}
+
+# Adds marker-cluster-plugin HTML dependency
+markerClusterDependencies <- function() {
   list(
     htmltools::htmlDependency(
       'leaflet-markercluster',
-      '0.4.0',
+      '0.5.0',
       system.file('htmlwidgets/plugins/Leaflet.markercluster', package = 'leaflet'),
-      script = 'leaflet.markercluster.js',
+      script = c('leaflet.markercluster.js', 'leaflet.markercluster.layersupport-src.js', 'leaflet.markercluster.freezable-src.js'),
       stylesheet = c('MarkerCluster.css', 'MarkerCluster.Default.css')
     )
   )
@@ -497,7 +610,7 @@ markerClusterDependencies = function() {
 #' )
 #'
 #' iconSet[c('red', 'green', 'red')]
-iconList = function(...) {
+iconList <- function(...) {
   res = structure(
     list(...),
     class = "leaflet_icon_set"
@@ -509,7 +622,7 @@ iconList = function(...) {
 }
 
 #' @export
-`[.leaflet_icon_set` = function(x, i) {
+`[.leaflet_icon_set` <- function(x, i) {
   if (is.factor(i)) {
     i = as.character(i)
   }
@@ -521,7 +634,7 @@ iconList = function(...) {
   structure(.subset(x, i), class = "leaflet_icon_set")
 }
 
-iconSetToIcons = function(x) {
+iconSetToIcons <- function(x) {
   # c("iconUrl", "iconRetinaUrl", ...)
   cols = names(formals(makeIcon))
   # list(iconUrl = "iconUrl", iconRetinaUrl = "iconRetinaUrl", ...)
@@ -550,7 +663,7 @@ iconSetToIcons = function(x) {
 #' @inheritParams icons
 #'
 #' @export
-makeIcon = function(iconUrl = NULL, iconRetinaUrl = NULL, iconWidth = NULL, iconHeight = NULL,
+makeIcon <- function(iconUrl = NULL, iconRetinaUrl = NULL, iconWidth = NULL, iconHeight = NULL,
   iconAnchorX = NULL, iconAnchorY = NULL, shadowUrl = NULL, shadowRetinaUrl = NULL,
   shadowWidth = NULL, shadowHeight = NULL, shadowAnchorX = NULL, shadowAnchorY = NULL,
   popupAnchorX = NULL, popupAnchorY = NULL, className = NULL) {
@@ -594,7 +707,7 @@ makeIcon = function(iconUrl = NULL, iconRetinaUrl = NULL, iconWidth = NULL, icon
 #'   \code{\link{addMarkers}()}.
 #' @export
 #' @example inst/examples/icons.R
-icons = function(
+icons <- function(
   iconUrl = NULL, iconRetinaUrl = NULL, iconWidth = NULL, iconHeight = NULL,
   iconAnchorX = NULL, iconAnchorY = NULL, shadowUrl = NULL, shadowRetinaUrl = NULL,
   shadowWidth = NULL, shadowHeight = NULL, shadowAnchorX = NULL, shadowAnchorY = NULL,
@@ -612,7 +725,7 @@ icons = function(
   ))
 }
 
-packStrings = function(strings) {
+packStrings <- function(strings) {
   if (length(strings) == 0) {
     return(NULL)
   }
@@ -626,7 +739,7 @@ packStrings = function(strings) {
   )
 }
 
-b64EncodePackedIcons = function(packedIcons) {
+b64EncodePackedIcons <- function(packedIcons) {
   if (is.null(packedIcons))
     return(packedIcons)
 
@@ -649,7 +762,7 @@ b64EncodePackedIcons = function(packedIcons) {
 #'   marker options; see \url{http://leafletjs.com/reference.html#marker}
 #' @describeIn map-options Options for markers
 #' @export
-markerOptions = function(
+markerOptions <- function(
   clickable = TRUE,
   draggable = FALSE,
   keyboard = TRUE,
@@ -658,12 +771,13 @@ markerOptions = function(
   zIndexOffset = 0,
   opacity = 1.0,
   riseOnHover = FALSE,
-  riseOffset = 250
+  riseOffset = 250,
+  ...
 ) {
   list(
     clickable = clickable, draggable = draggable, keyboard = keyboard,
     title = title, alt = alt, zIndexOffset = zIndexOffset, opacity = opacity,
-    riseOnHover = riseOnHover, riseOffset = riseOffset
+    riseOnHover = riseOnHover, riseOffset = riseOffset, ...
   )
 }
 
@@ -674,17 +788,26 @@ markerOptions = function(
 #'   spiderfy it so you can see all of its markers
 #' @param removeOutsideVisibleBounds clusters and markers too far from the
 #'   viewport are removed from the map for performance
+#' @param spiderLegPolylineOptions Allows you to specify PolylineOptions (\url{http://leafletjs.com/reference.html#polyline-options}) to style spider legs. By default, they are { weight: 1.5, color: '#222', opacity: 0.5 }
+#' @param freezeAtZoom Allows you to freeze cluster expansion to a zoom level.
+#' Can be a zoom level e.g. 10, 12 or 'max' or 'maxKeepSpiderify'
+#' See \url{https://github.com/ghybs/Leaflet.MarkerCluster.Freezable#api-reference}
 #' @describeIn map-options Options for marker clusters
 #' @export
-markerClusterOptions = function(
+markerClusterOptions <- function(
   showCoverageOnHover = TRUE, zoomToBoundsOnClick = TRUE,
-  spiderfyOnMaxZoom = TRUE, removeOutsideVisibleBounds = TRUE, ...
+  spiderfyOnMaxZoom = TRUE, removeOutsideVisibleBounds = TRUE,
+  spiderLegPolylineOptions = list(weight= 1.5, color= '#222', opacity= 0.5),
+  freezeAtZoom = FALSE,
+  ...
 ) {
   list(
     showCoverageOnHover = showCoverageOnHover,
     zoomToBoundsOnClick = zoomToBoundsOnClick,
     spiderfyOnMaxZoom = spiderfyOnMaxZoom,
-    removeOutsideVisibleBounds = removeOutsideVisibleBounds, ...
+    removeOutsideVisibleBounds = removeOutsideVisibleBounds,
+    spiderLegPolylineOptions =  spiderLegPolylineOptions,
+    freezeAtZoom = freezeAtZoom, ...
   )
 }
 
@@ -705,7 +828,7 @@ markerClusterOptions = function(
 #'   pattern}
 #' @describeIn map-layers Add circle markers to the map
 #' @export
-addCircleMarkers = function(
+addCircleMarkers <- function(
   map, lng = NULL, lat = NULL, radius = 10, layerId = NULL, group = NULL,
   stroke = TRUE,
   color = "#03F",
@@ -716,6 +839,9 @@ addCircleMarkers = function(
   fillOpacity = 0.2,
   dashArray = NULL,
   popup = NULL,
+  popupOptions = NULL,
+  label = NULL,
+  labelOptions = NULL,
   options = pathOptions(),
   clusterOptions = NULL,
   clusterId = NULL,
@@ -730,38 +856,40 @@ addCircleMarkers = function(
     map$dependencies = c(map$dependencies, markerClusterDependencies())
   pts = derivePoints(data, lng, lat, missing(lng), missing(lat), "addCircleMarkers")
   invokeMethod(map, data, 'addCircleMarkers', pts$lat, pts$lng, radius,
-      layerId, group, options, clusterOptions, clusterId, popup) %>%
+               layerId, group, options, clusterOptions, clusterId,
+               popup, popupOptions, safeLabel(label, data), labelOptions,
+               getCrosstalkOptions(data)) %>%
     expandLimits(pts$lat, pts$lng)
 }
 
 #' @rdname remove
 #' @export
-removeMarker = function(map, layerId) {
+removeMarker <- function(map, layerId) {
   invokeMethod(map, getMapData(map), 'removeMarker', layerId)
 }
 
 #' @rdname remove
 #' @export
-clearMarkers = function(map) {
+clearMarkers <- function(map) {
   invokeMethod(map, NULL, 'clearMarkers')
 }
 
 #' @rdname remove
 #' @export
-removeMarkerCluster = function(map, layerId) {
+removeMarkerCluster <- function(map, layerId) {
   invokeMethod(map, getMapData(map), 'removeMarkerCluster', layerId)
 }
 
 #' @rdname remove
 #' @export
-clearMarkerClusters = function(map) {
+clearMarkerClusters <- function(map) {
   invokeMethod(map, NULL, 'clearMarkerClusters')
 }
 
 #' @param clusterId the id of the marker cluster layer
 #' @rdname remove
 #' @export
-removeMarkerFromCluster = function(map, layerId, clusterId) {
+removeMarkerFromCluster <- function(map, layerId, clusterId) {
   invokeMethod(map, getMapData(map), 'removeMarkerFromCluster', layerId, clusterId)
 }
 
@@ -776,22 +904,54 @@ removeMarkerFromCluster = function(map, layerId, clusterId) {
 #' @describeIn map-options Options for vector layers (polylines, polygons,
 #'   rectangles, and circles, etc)
 #' @export
-pathOptions = function(
+pathOptions <- function(
   lineCap = NULL,
   lineJoin = NULL,
   clickable = TRUE,
   pointerEvents = NULL,
-  className = ""
+  className = "",
+  ...
 ) {
   list(
     lineCap = lineCap, lineJoin = lineJoin, clickable = clickable,
-    pointerEvents = pointerEvents, className = className
+    pointerEvents = pointerEvents, className = className, ...
   )
 }
 
+#' Options to highlight shapes (polylines/polygons/circles/rectangles)
+#' @param bringToFront Whether the shape should be brought to front on hover.
+#' @param sendToBack whether the shape should be sent to back on mouse out.
+#' @describeIn map-layers Options to highligh a shape on hover
+#' @export
+highlightOptions <- function(
+  stroke = NULL,
+  color = NULL,
+  weight = NULL,
+  opacity = NULL,
+  fill = NULL,
+  fillColor = NULL,
+  fillOpacity = NULL,
+  dashArray = NULL,
+  bringToFront = NULL,
+  sendToBack = NULL
+) {
+  filterNULL(list(
+    stroke = stroke,
+    color = color,
+    weight = weight,
+    opacity = opacity,
+    fill = fill,
+    fillColor = fillColor,
+    fillOpacity = fillOpacity,
+    dashArray = dashArray,
+    bringToFront = bringToFront,
+    sendToBack = sendToBack
+  ))
+}
+
 #' @describeIn map-layers Add circles to the map
 #' @export
-addCircles = function(
+addCircles <- function(
   map, lng = NULL, lat = NULL, radius = 10, layerId = NULL, group = NULL,
   stroke = TRUE,
   color = "#03F",
@@ -802,7 +962,11 @@ addCircles = function(
   fillOpacity = 0.2,
   dashArray = NULL,
   popup = NULL,
+  popupOptions = NULL,
+  label = NULL,
+  labelOptions = NULL,
   options = pathOptions(),
+  highlightOptions = NULL,
   data = getMapData(map)
 ) {
   options = c(options, list(
@@ -811,16 +975,19 @@ addCircles = function(
     dashArray = dashArray
   ))
   pts = derivePoints(data, lng, lat, missing(lng), missing(lat), "addCircles")
-  invokeMethod(map, data, 'addCircles', pts$lat, pts$lng, radius, layerId, group, options, popup) %>%
+  invokeMethod(map, data, 'addCircles', pts$lat, pts$lng, radius, layerId, group, options,
+               popup, popupOptions, safeLabel(label, data), labelOptions, highlightOptions,
+               getCrosstalkOptions(data)) %>%
     expandLimits(pts$lat, pts$lng)
 }
 
 #' @param smoothFactor how much to simplify the polyline on each zoom level
 #'   (more means better performance and less accurate representation)
+#' @param highlightOptions Options for highlighting the shape on mouse over.
 #' @param noClip whether to disable polyline clipping
 #' @describeIn map-layers Add polylines to the map
 #' @export
-addPolylines = function(
+addPolylines <- function(
   map, lng = NULL, lat = NULL, layerId = NULL, group = NULL,
   stroke = TRUE,
   color = "#03F",
@@ -833,7 +1000,11 @@ addPolylines = function(
   smoothFactor = 1.0,
   noClip = FALSE,
   popup = NULL,
+  popupOptions = NULL,
+  label = NULL,
+  labelOptions = NULL,
   options = pathOptions(),
+  highlightOptions = NULL,
   data = getMapData(map)
 ) {
   options = c(options, list(
@@ -842,7 +1013,8 @@ addPolylines = function(
     dashArray = dashArray, smoothFactor = smoothFactor, noClip = noClip
   ))
   pgons = derivePolygons(data, lng, lat, missing(lng), missing(lat), "addPolylines")
-  invokeMethod(map, data, 'addPolylines', pgons, layerId, group, options, popup) %>%
+  invokeMethod(map, data, 'addPolylines', pgons, layerId, group, options,
+               popup, popupOptions, safeLabel(label, data), labelOptions, highlightOptions) %>%
     expandLimitsBbox(pgons)
 }
 
@@ -850,7 +1022,7 @@ addPolylines = function(
 #'   north-east corners of rectangles
 #' @describeIn map-layers Add rectangles to the map
 #' @export
-addRectangles = function(
+addRectangles <- function(
   map, lng1, lat1, lng2, lat2, layerId = NULL, group = NULL,
   stroke = TRUE,
   color = "#03F",
@@ -863,7 +1035,11 @@ addRectangles = function(
   smoothFactor = 1.0,
   noClip = FALSE,
   popup = NULL,
+  popupOptions = NULL,
+  label = NULL,
+  labelOptions = NULL,
   options = pathOptions(),
+  highlightOptions = NULL,
   data = getMapData(map)
 ) {
   options = c(options, list(
@@ -875,13 +1051,17 @@ addRectangles = function(
   lat1 = resolveFormula(lat1, data)
   lng2 = resolveFormula(lng2, data)
   lat2 = resolveFormula(lat2, data)
-  invokeMethod(map, data, 'addRectangles',lat1, lng1, lat2, lng2, layerId, group, options, popup) %>%
+
+  df1 <- validateCoords(lng1, lat1, "addRectangles")
+  df2 <- validateCoords(lng2, lat2, "addRectangles")
+
+  invokeMethod(map, data, 'addRectangles',df1$lat, df1$lng, df2$lat, df2$lng, layerId, group, options, popup, popupOptions, safeLabel(label, data), labelOptions, highlightOptions) %>%
     expandLimits(c(lat1, lat2), c(lng1, lng2))
 }
 
 #' @describeIn map-layers Add polygons to the map
 #' @export
-addPolygons = function(
+addPolygons <- function(
   map, lng = NULL, lat = NULL, layerId = NULL, group = NULL,
   stroke = TRUE,
   color = "#03F",
@@ -894,7 +1074,11 @@ addPolygons = function(
   smoothFactor = 1.0,
   noClip = FALSE,
   popup = NULL,
+  popupOptions = NULL,
+  label = NULL,
+  labelOptions = NULL,
   options = pathOptions(),
+  highlightOptions = NULL,
   data = getMapData(map)
 ) {
   options = c(options, list(
@@ -903,26 +1087,26 @@ addPolygons = function(
     dashArray = dashArray, smoothFactor = smoothFactor, noClip = noClip
   ))
   pgons = derivePolygons(data, lng, lat, missing(lng), missing(lat), "addPolygons")
-  invokeMethod(map, data, 'addPolygons', pgons, layerId, group, options, popup) %>%
+  invokeMethod(map, data, 'addPolygons', pgons, layerId, group, options, popup, popupOptions, safeLabel(label, data), labelOptions, highlightOptions) %>%
     expandLimitsBbox(pgons)
 }
 
 #' @rdname remove
 #' @export
-removeShape = function(map, layerId) {
+removeShape <- function(map, layerId) {
   invokeMethod(map, getMapData(map), 'removeShape', layerId)
 }
 
 #' @rdname remove
 #' @export
-clearShapes = function(map) {
+clearShapes <- function(map) {
   invokeMethod(map, NULL, 'clearShapes')
 }
 
 #' @param geojson a GeoJSON list, or character vector of length 1
 #' @describeIn map-layers Add GeoJSON layers to the map
 #' @export
-addGeoJSON = function(map, geojson, layerId = NULL, group = NULL,
+addGeoJSON <- function(map, geojson, layerId = NULL, group = NULL,
   stroke = TRUE,
   color = "#03F",
   weight = 5,
@@ -945,13 +1129,13 @@ addGeoJSON = function(map, geojson, layerId = NULL, group = NULL,
 
 #' @rdname remove
 #' @export
-removeGeoJSON = function(map, layerId) {
+removeGeoJSON <- function(map, layerId) {
   invokeMethod(map, getMapData(map), 'removeGeoJSON', layerId)
 }
 
 #' @rdname remove
 #' @export
-clearGeoJSON = function(map) {
+clearGeoJSON <- function(map) {
   invokeMethod(map, NULL, 'clearGeoJSON')
 }
 
@@ -986,7 +1170,7 @@ clearGeoJSON = function(map) {
 #' }
 #'
 #' @export
-addLayersControl = function(map,
+addLayersControl <- function(map,
   baseGroups = character(0), overlayGroups = character(0),
   position = c('topright', 'bottomright', 'bottomleft', 'topleft'),
   options = layersControlOptions()) {
@@ -1002,13 +1186,14 @@ addLayersControl = function(map,
 #'   to have the layers control always appear in its expanded state.
 #' @param autoZIndex if \code{TRUE}, the control will automatically maintain
 #'   the z-order of its various groups as overlays are switched on and off.
+#' @param ... other options for \code{layersControlOptions()}
 #' @export
-layersControlOptions = function(collapsed = TRUE, autoZIndex = TRUE) {
-  list(collapsed = collapsed, autoZIndex = autoZIndex)
+layersControlOptions <- function(collapsed = TRUE, autoZIndex = TRUE, ...) {
+  list(collapsed = collapsed, autoZIndex = autoZIndex, ...)
 }
 
 #' @rdname addLayersControl
 #' @export
-removeLayersControl = function(map) {
+removeLayersControl <- function(map) {
   invokeMethod(map, NULL, 'removeLayersControl')
 }
