@@ -2,7 +2,7 @@ leafletOmnivoreDependencies <- function() {
   list(
     htmltools::htmlDependency(
       "leaflet-omnivore",
-      "0.3.0",
+      "0.3.3",
       system.file("htmlwidgets/lib/leaflet-omnivore", package = "leaflet"),
       script = "leaflet-omnivore.min.js"
     )
@@ -13,7 +13,7 @@ leafletOmnivoreDependencies <- function() {
 #' @param topojson a TopoJSON list, or character vector of length 1
 #' @describeIn map-layers Add TopoJSON layers to the map
 #' @export
-addTopoJSON = function(map, topojson, layerId = NULL, group = NULL,
+addTopoJSON <- function(map, topojson, layerId = NULL, group = NULL,
   stroke = TRUE,
   color = "#03F",
   weight = 5,
@@ -34,3 +34,15 @@ addTopoJSON = function(map, topojson, layerId = NULL, group = NULL,
   ))
   invokeMethod(map, getMapData(map), 'addTopoJSON', topojson, layerId, group, options)
 }
+
+#' @rdname remove
+#' @export
+removeTopoJSON <- function(map, layerId) {
+  invokeMethod(map, getMapData(map), 'removeTopoJSON', layerId)
+}
+
+#' @rdname remove
+#' @export
+clearTopoJSON <- function(map) {
+  invokeMethod(map, NULL, 'clearTopoJSON')
+}
