@@ -19,13 +19,13 @@
 #' )
 #'
 #' if (interactive()) print(app)}
-leafletOutput = function(outputId, width = "100%", height = 400) {
+leafletOutput <- function(outputId, width = "100%", height = 400) {
   htmlwidgets::shinyWidgetOutput(outputId, "leaflet", width, height, "leaflet")
 }
 
 #' @rdname map-shiny
 #' @export
-renderLeaflet = function(expr, env = parent.frame(), quoted = FALSE) {
+renderLeaflet <- function(expr, env = parent.frame(), quoted = FALSE) {
   if (!quoted) expr = substitute(expr)  # force quoted
   htmlwidgets::shinyRenderWidget(expr, leafletOutput, env, quoted = TRUE)
 }
